@@ -2,16 +2,27 @@ from __future__ import annotations
 
 from functools import total_ordering
 from itertools import product
-from typing import Hashable, FrozenSet, Callable, Optional, Union
+from typing import Hashable, FrozenSet, Callable, Union
 
 import attr
 import funcy as fn
 
 
+def _freeze(alphabet):
+    if alphabet is None or isinstance(alphabet, SupAlphabet):
+        return SupAlphabet()
+    elif isinstance(alphabet, ProductAlphabet):
+        return alphabet
+    return frozenset(alphabet)
+
+
 @attr.s(frozen=True, auto_attribs=True, eq=False)
 @total_ordering
 class SupAlphabet:
     """Alphabet containing all other alphabets."""
+    def __hash__(self):
+        return hash(id(self))
+
     def __contains__(self, _):
         return True
 
@@ -21,16 +32,16 @@ class SupAlphabet:
     def __eq__(self, other):
         return isinstance(other, SupAlphabet)
 
-    def __len__(self):
-        raise NotImplementedError("SupAlphabet has no defined length.")
-
 
 @attr.s(frozen=True, auto_attribs=True, repr=False, eq=False)
 @total_ordering
 class ProductAlphabet:
     """Implicit encoding of product alphabet."""
-    left: Alphabet
-    right: Alphabet
+    left: Alphabet = attr.ib(converter=_freeze)
+    right: Alphabet = attr.ib(converter=_freeze)
+
+    def __hash__(self):
+        return hash((self.left, self.right))
 
     def __eq__(self, other):
         return (self.left == other.left) and (self.right == other.right)
@@ -61,14 +72,6 @@ Letter = Hashable
 Alphabet = Union[FrozenSet[Letter], ProductAlphabet, SupAlphabet]
 
 
-def _freeze(alphabet: Optional[Alphabet] = None):
-    if alphabet is None or isinstance(alphabet, SupAlphabet):
-        return SupAlphabet()
-    elif isinstance(alphabet, ProductAlphabet):
-        return alphabet
-    return frozenset(alphabet)
-
-
 @attr.s(frozen=True, auto_attribs=True)
 class DFA:
     """Represents a Discrete Finite Automaton or Moore Machine."""
